import asyncio
import logging
from collections import OrderedDict
from itertools import groupby
from pathlib import Path
from typing import Tuple, Dict, Optional, List

import aiosql
from asyncpg.pool import Pool
from migra import Statements, Changes
from schemainspect import ColumnInfo
from schemainspect.misc import quoted_identifier
from schemainspect.pg.obj import (
    InspectedSchema,
    InspectedRowPolicy,
    InspectedCollation,
    InspectedPrivilege,
    InspectedEnum,
    InspectedSelectable,
    InspectedIndex,
    InspectedSequence,
    InspectedConstraint,
    InspectedExtension,
    InspectedFunction,
    InspectedTrigger,
    InspectedDomain,
    InspectedType,
)

from aiosqlembic.models.inspector import (
    Schema,
    Enum,
    Relation,
    Index,
    Sequence,
    Constraint,
    Extension,
    Function,
    Dep,
    Privilege,
    Trigger,
    Collation,
    RlsPolicy,
    Type,
    Domain,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logging.basicConfig(
    format="%(levelname)s:%(filename)s.%(funcName)s:%(lineno)d %(message)s"
)


class AsyncSchemaInspector:
    def __init__(self, pool: Pool):
        self.current_path = Path(__file__).parent
        self.queries = aiosql.from_path(
            self.current_path / "./sql/metadata",
            "asyncpg",
            record_classes={
                "Schema": Schema,
                "Enum": Enum,
                "Relation": Relation,
                "Index": Index,
                "Sequence": Sequence,
                "Constraint": Constraint,
                "Extension": Extension,
                "Function": Function,
                "Dep": Dep,
                "Privilege": Privilege,
                "Trigger": Trigger,
                "Collation": Collation,
                "RlsPolicy": RlsPolicy,
                "Type": Type,
                "Domain": Domain,
            },
        )
        self.pool = pool

    async def load_all(self) -> None:
        task_schemas = asyncio.create_task(self.load_schemas())
        task_relations = asyncio.create_task(self.load_all_relations())
        task_functions = asyncio.create_task(self.load_functions())
        await asyncio.gather(task_schemas, task_relations, task_functions)
        self.selectables: OrderedDict[str, InspectedSelectable] = OrderedDict()
        self.selectables.update(self.relations)
        self.selectables.update(self.functions)

        task_deps = asyncio.create_task(self.load_deps())
        task_deps_all = asyncio.create_task(self.load_deps_all())
        task_privileges = asyncio.create_task(self.load_privileges())
        task_triggers = asyncio.create_task(self.load_triggers())
        task_collations = asyncio.create_task(self.load_collations())
        task_rlspolicies = asyncio.create_task(self.load_rlspolicies())
        task_types = asyncio.create_task(self.load_types())
        task_domains = asyncio.create_task(self.load_domains())
        await asyncio.gather(
            task_deps,
            task_deps_all,
            task_privileges,
            task_triggers,
            task_collations,
            task_rlspolicies,
            task_types,
            task_domains,
        )
        pass

    async def load_schemas(self) -> None:
        q: List[Schema] = await self.queries.get_all_schemas(self.pool)
        logger.debug(f"schemas loaded: {q}")

        schemas: List[InspectedSchema] = [
            InspectedSchema(schema=each.schemaname) for each in q
        ]
        self.schemas: OrderedDict[str, InspectedSchema] = OrderedDict(
            (schema.schema, schema) for schema in schemas
        )

    async def load_all_relations(self) -> None:
        self.tables: OrderedDict[str, InspectedSelectable] = OrderedDict()
        self.views: OrderedDict[str, InspectedSelectable] = OrderedDict()
        self.materialized_views: OrderedDict[str, InspectedSelectable] = OrderedDict()
        self.composite_types: OrderedDict[str, InspectedSelectable] = OrderedDict()

        q = await self.queries.get_all_enums(self.pool)
        logger.debug(f"enums loaded: {q}")

        enumlist = [
            InspectedEnum(name=i.name, schema=i.schemaname, elements=i.elements)
            for i in q
        ]
        self.enums: OrderedDict[str, InspectedEnum] = OrderedDict(
            (i.quoted_full_name, i) for i in enumlist
        )
        q = await self.queries.get_all_relations(self.pool)
        logger.debug(f"relations loaded: {q}")

        for _, g in groupby(q, lambda x: (x.relationtype, x.schemaname, x.name)):
            clist = list(g)
            f = clist[0]

            def get_enum(name: str, schema: str) -> Optional[InspectedEnum]:
                if not name and not schema:
                    return None

                quoted_full_name = "{}.{}".format(
                    quoted_identifier(schema), quoted_identifier(name)
                )
                return self.enums[quoted_full_name]

            columns = [
                ColumnInfo(
                    name=c.attname,
                    dbtype=c.datatype,
                    dbtypestr=c.datatypestring,
                    # pytype=self.to_pytype(c.datatype),
                    pytype=c.datatype,
                    default=c.defaultdef,
                    not_null=c.not_null,
                    is_enum=c.is_enum,
                    enum=get_enum(c.enum_name, c.enum_schema),
                    collation=c.collation,
                )
                for c in clist
                if c.position_number
            ]

            s = InspectedSelectable(
                name=f.name,
                schema=f.schemaname,
                columns=OrderedDict((c.name, c) for c in columns),
                relationtype=f.relationtype,
                definition=f.definition,
                comment=f.comment,
                parent_table=f.parent_table,
                partition_def=f.partition_def,
                rowsecurity=f.rowsecurity,
                forcerowsecurity=f.forcerowsecurity,
            )
            RELATIONTYPES = {
                "r": "tables",
                "v": "views",
                "m": "materialized_views",
                "c": "composite_types",
                "p": "tables",
            }
            att = getattr(self, RELATIONTYPES[f.relationtype])
            att[s.quoted_full_name] = s
        self.relations: OrderedDict[str, InspectedSelectable] = OrderedDict()
        for x in (self.tables, self.views, self.materialized_views):
            self.relations.update(x)
        q = await self.queries.get_all_indexes(self.pool)
        logger.debug(f"indexes loaded: {q}")

        indexlist = [
            InspectedIndex(
                name=i.name,
                schema=i.schemaname,
                definition=i.definition,
                table_name=i.table_name,
                key_columns=i.key_columns,
                key_options=i.key_options,
                num_att=i.num_att,
                is_unique=i.is_unique,
                is_pk=i.is_pk,
                is_exclusion=i.is_exclusion,
                is_immediate=i.is_immediate,
                is_clustered=i.is_clustered,
                key_collations=i.key_collations,
                key_expressions=i.key_expressions,
                partial_predicate=i.partial_predicate,
            )
            for i in q
        ]
        self.indexes: OrderedDict[str, InspectedIndex] = OrderedDict(
            (i.quoted_full_name, i) for i in indexlist
        )
        q = await self.queries.get_all_sequences(self.pool)
        logger.debug(f"sequences loaded: {q}")

        sequencelist = [InspectedSequence(name=i.name, schema=i.schemaname) for i in q]
        self.sequences: OrderedDict[str, InspectedSequence] = OrderedDict(
            (i.quoted_full_name, i) for i in sequencelist
        )
        q = await self.queries.get_all_constraints(self.pool)

        constraintlist = []

        for i in q:
            constraint = InspectedConstraint(
                name=i.name,
                schema=i.schemaname,
                constraint_type=i.constraint_type,
                table_name=i.table_name,
                definition=i.definition,
                index=i.index,
            )
            if constraint.index:
                index_name = quoted_identifier(i.index, schema=i.schemaname)
                index = self.indexes[index_name]
                index.constraint = constraint
                constraint.index = index

            constraintlist.append(constraint)

        self.constraints: OrderedDict[str, InspectedConstraint] = OrderedDict(
            (i.quoted_full_name, i) for i in constraintlist
        )

        q = await self.queries.get_all_extensions(self.pool)
        logger.debug(f"extensions loaded: {q}")

        extensionlist = [
            InspectedExtension(name=i.name, schema=i.schemaname, version=i.version)
            for i in q
        ]
        # extension names are unique per-database rather than per-schema like other things (even though extensions are assigned to a particular schema)
        self.extensions: OrderedDict[str, InspectedExtension] = OrderedDict(
            (i.name, i) for i in extensionlist
        )
        # add indexes and constraints to each table
        for each in self.indexes.values():
            t = each.quoted_full_table_name
            n = each.quoted_full_name
            self.relations[t].indexes[n] = each
        for each in self.constraints.values():
            t = each.quoted_full_table_name
            n = each.quoted_full_name
            self.relations[t].constraints[n] = each

    async def load_functions(self) -> None:
        self.functions = OrderedDict()
        q = await self.queries.get_all_functions(self.pool)
        logger.debug(f"functions loaded: {q}")

        for _, g in groupby(q, lambda x: (x.schemaname, x.name, x.identity_arguments)):
            clist = list(g)
            f = clist[0]
            outs = [c for c in clist if c.parameter_mode == "OUT"]
            columns = [
                ColumnInfo(
                    name=c.parameter_name,
                    dbtype=c.data_type,
                    pytype=c.data_type,
                    # pytype=self.to_pytype(c.data_type),
                )
                for c in outs
            ]
            if outs:
                columns = [
                    ColumnInfo(
                        name=c.parameter_name,
                        dbtype=c.data_type,
                        pytype=c.data_type,
                        # pytype=self.to_pytype(c.data_type),
                    )
                    for c in outs
                ]
            else:
                columns = [
                    ColumnInfo(
                        name=f.name,
                        dbtype=f.data_type,
                        pytype=f.returntype,
                        # pytype=self.to_pytype(f.returntype),
                        default=f.parameter_default,
                    )
                ]
            plist = [
                ColumnInfo(
                    name=c.parameter_name,
                    dbtype=c.data_type,
                    pytype=c.data_type,
                    # pytype=self.to_pytype(c.data_type),
                    default=c.parameter_default,
                )
                for c in clist
                if c.parameter_mode == "IN"
            ]
            s = InspectedFunction(
                schema=f.schemaname,
                name=f.name,
                columns=OrderedDict((c.name, c) for c in columns),
                inputs=plist,
                identity_arguments=f.identity_arguments,
                result_string=f.result_string,
                language=f.language,
                definition=f.definition,
                strictness=f.strictness,
                security_type=f.security_type,
                volatility=f.volatility,
                full_definition=f.full_definition,
                comment=f.comment,
                returntype=f.returntype,
            )

            identity_arguments = "({})".format(s.identity_arguments)
            self.functions[s.quoted_full_name + identity_arguments] = s

    async def load_deps(self) -> None:
        q = await self.queries.get_all_deps(self.pool)
        logger.debug(f"deps loaded: {q}")

        for dep in q:
            x = quoted_identifier(dep.name, dep.schemaname)
            x_dependent_on = quoted_identifier(
                dep.name_dependent_on,
                dep.schema_dependent_on,
                dep.identity_arguments_dependent_on,
            )
            self.selectables[x].dependent_on.append(x_dependent_on)
            self.selectables[x].dependent_on.sort()
            self.selectables[x_dependent_on].dependents.append(x)
            self.selectables[x_dependent_on].dependents.sort()

    async def load_deps_all(self) -> None:
        def get_related_for_item(
            item: InspectedSelectable, att: str
        ) -> List[InspectedSelectable]:
            related = [self.selectables[_] for _ in getattr(item, att)]
            return [item.signature] + [
                _ for d in related for _ in get_related_for_item(d, att)
            ]

        for k, x in self.selectables.items():
            d_all = get_related_for_item(x, "dependent_on")[1:]
            d_all.sort()
            x.dependent_on_all = d_all
            d_all = get_related_for_item(x, "dependents")[1:]
            d_all.sort()
            x.dependents_all = d_all

    async def load_privileges(self) -> None:
        q = await self.queries.get_all_privileges(self.pool)
        logger.debug(f"privileges loaded: {q}")

        privileges = [
            InspectedPrivilege(
                object_type=i.object_type,
                schema=i.schemaname,
                name=i.name,
                privilege=i.privilege,
                target_user=i.user,
            )
            for i in q
        ]
        self.privileges: OrderedDict[str, InspectedPrivilege] = OrderedDict(
            (i.key, i) for i in privileges
        )

    async def load_triggers(self) -> None:
        q = await self.queries.get_all_triggers(self.pool)
        logger.debug(f"triggers loaded: {q}")

        triggers = [
            InspectedTrigger(
                i.name,
                i.schemaname,
                i.table_name,
                i.proc_schema,
                i.proc_name,
                i.enabled,
                i.full_definition,
            )
            for i in q
        ]
        self.triggers = OrderedDict((t.signature, t) for t in triggers)

    async def load_collations(self) -> None:
        q = await self.queries.get_all_collations(self.pool)
        logger.debug(f"collations loaded: {q}")

        collations = [
            InspectedCollation(
                schema=i.schemaname,
                name=i.name,
                provider=i.provider,
                encoding=i.encoding,
                lc_collate=i.lc_collate,
                lc_ctype=i.lc_ctype,
                version=i.version,
            )
            for i in q
        ]
        self.collations: OrderedDict[str, InspectedCollation] = OrderedDict(
            (i.quoted_full_name, i) for i in collations
        )

    async def load_rlspolicies(self) -> None:
        q = await self.queries.get_all_rlspolicies(self.pool)
        logger.debug(f"rlspolicies loaded: {q}")
        rlspolicies = [
            InspectedRowPolicy(
                name=p.name,
                schema=p.schemaname,
                table_name=p.table_name,
                commandtype=p.commandtype,
                permissive=p.permissive,
                roles=p.roles,
                qual=p.qual,
                withcheck=p.withcheck,
            )
            for p in q
        ]

        self.rlspolicies: OrderedDict[str, InspectedRowPolicy] = OrderedDict(
            (p.key, p) for p in rlspolicies
        )

    async def load_types(self) -> None:
        q = await self.queries.get_all_types(self.pool)
        logger.debug(f"types loaded: {q}")

        def col(defn: Dict[str, str]) -> Tuple[str, str]:
            return defn["attribute"], defn["type"]

        types = [
            InspectedType(i.name, i.schemaname, dict(col(_) for _ in i.columns))
            for i in q
        ]
        self.types = OrderedDict((t.signature, t) for t in types)

    async def load_domains(self) -> None:
        q = await self.queries.get_all_domains(self.pool)
        logger.debug(f"domains loaded: {q}")

        def col(defn: Dict[str, str]) -> Tuple[str, str]:
            return defn["attribute"], defn["type"]

        domains = [
            InspectedDomain(
                i.name,
                i.schemaname,
                i.data_type,
                i.collation,
                i.constraint_name,
                i.not_null,
                i.default,
                i.check,
            )
            for i in q
        ]
        self.domains = OrderedDict((t.signature, t) for t in domains)

    # @property
    # def partitioned_tables(self) -> "OrderedDict[str, InspectedSelectable]":
    #     return OrderedDict((k, v) for k, v in self.tables.items() if v.is_partitioned)
    #
    # @property
    # def alterable_tables(
    #     self,
    # ) -> "OrderedDict[str, InspectedSelectable]":  # ordinary tables and parent tables
    #     return OrderedDict((k, v) for k, v in self.tables.items() if v.is_alterable)
    #
    # @property
    # def data_tables(
    #     self,
    # ) -> "OrderedDict[str, InspectedSelectable]":  # ordinary tables and child tables
    #     return OrderedDict((k, v) for k, v in self.tables.items() if v.contains_data)
    #
    # @property
    # def partitioning_child_tables(self) -> "OrderedDict[str, InspectedSelectable]":
    #     return OrderedDict(
    #         (k, v) for k, v in self.tables.items() if v.is_partitioning_child_table
    #     )
    #
    # @property
    # def tables_using_partitioning(self) -> "OrderedDict[str, InspectedSelectable]":
    #     return OrderedDict(
    #         (k, v) for k, v in self.tables.items() if v.uses_partitioning
    #     )
    #
    # @property
    # def tables_not_using_partitioning(self) -> "OrderedDict[str, InspectedSelectable]":
    #     return OrderedDict(
    #         (k, v) for k, v in self.tables.items() if not v.uses_partitioning
    #     )

    def one_schema(self, schema: str) -> None:
        props = "schemas relations tables views functions selectables sequences constraints indexes enums extensions privileges collations triggers"
        for prop in props.split():
            att = getattr(self, prop)
            filtered = {k: v for k, v in att.items() if v.schema == schema}
            setattr(self, prop, filtered)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AsyncSchemaInspector):
            return NotImplemented
        else:
            return (
                type(self) == type(other)
                and self.schemas == other.schemas
                and self.relations == other.relations
                and self.sequences == other.sequences
                and self.enums == other.enums
                and self.constraints == other.constraints
                and self.extensions == other.extensions
                and self.functions == other.functions
                and self.triggers == other.triggers
                and self.collations == other.collations
                and self.rlspolicies == other.rlspolicies
            )


class Amigration:
    def __init__(self, schema: str = None) -> None:
        self.statements = Statements()
        self.changes = Changes(None, None)
        self.schema = schema

    @classmethod
    async def create(cls, x_from: Pool, x_target: Pool) -> "Amigration":
        self = Amigration()
        self.changes.i_from = AsyncSchemaInspector(pool=x_from)
        self.changes.i_target = AsyncSchemaInspector(pool=x_target)
        await self.load_all()
        return self

    async def load_all(self) -> None:
        logger.debug("loading migration")
        task_ifrom = self.changes.i_from.load_all()
        task_itarget = self.changes.i_target.load_all()
        await asyncio.gather(task_ifrom, task_itarget)

    def clear(self) -> None:
        self.statements = Statements()

    # def apply(self):
    #     for stmt in self.statements:
    #         raw_execute(self.s_from, stmt)
    #     self.changes.i_from = get_inspector(self.s_from, schema=self.schema)
    #     safety_on = self.statements.safe
    #     self.clear()
    #     self.set_safety(safety_on)

    def add(self, statements: Statements) -> None:
        self.statements += statements

    def add_sql(self, sql: str) -> None:
        self.statements += Statements([sql])

    def set_safety(self, safety_on: bool) -> None:
        self.statements.safe = safety_on

    def add_extension_changes(self, creates: bool = True, drops: bool = True) -> None:
        if creates:
            self.add(self.changes.extensions(creations_only=True))
        if drops:
            self.add(self.changes.extensions(drops_only=True))

    def add_all_changes(self, privileges: bool = False) -> None:
        self.add(self.changes.schemas(creations_only=True))

        self.add(self.changes.extensions(creations_only=True))
        self.add(self.changes.collations(creations_only=True))
        self.add(self.changes.enums(creations_only=True, modifications=False))
        self.add(self.changes.sequences(creations_only=True))
        self.add(self.changes.triggers(drops_only=True))
        self.add(self.changes.rlspolicies(drops_only=True))
        if privileges:
            self.add(self.changes.privileges(drops_only=True))
        self.add(self.changes.non_pk_constraints(drops_only=True))
        self.add(self.changes.pk_constraints(drops_only=True))
        self.add(self.changes.indexes(drops_only=True))

        self.add(self.changes.selectables())

        self.add(self.changes.sequences(drops_only=True))
        self.add(self.changes.enums(drops_only=True, modifications=False))
        self.add(self.changes.extensions(drops_only=True))
        self.add(self.changes.indexes(creations_only=True))
        self.add(self.changes.pk_constraints(creations_only=True))
        self.add(self.changes.non_pk_constraints(creations_only=True))
        if privileges:
            self.add(self.changes.privileges(creations_only=True))
        self.add(self.changes.rlspolicies(creations_only=True))
        self.add(self.changes.triggers(creations_only=True))
        self.add(self.changes.collations(drops_only=True))
        self.add(self.changes.schemas(drops_only=True))

    @property
    def sql(self) -> str:
        s: str = self.statements.sql
        return s



from datetime import datetime
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Union, Any, List

import aiosql
import asyncpg
from aiosql.aiosql import _ADAPTERS
import asyncclick as click
from aiosql.queries import Queries
from click import Context

from aiosqlembic.models.cli import (
    ConDB,
    ConDBTable,
    Version,
    MigrateRevision,
)
from aiosqlembic.schema import Amigration
from aiosqlembic.utils import (
    get_revision_tree,
    print_version,
    get_connection,
    camel2snake,
    templateEnv,
    create_test_database,
    get_next_revision,
    get_current_revision,
)

MIN_VERSION = 0
MAX_VERSION = 2 ** (64 - 1) - 1  # ie max size of sqlite3 integer and postgresql bigint

logging.basicConfig(
    format="%(levelname)s:%(filename)s.%(funcName)s:%(lineno)d %(message)s"
)
logger = logging.getLogger(__name__)


@dataclass
class AiosqlembicContext:
    debug: bool
    migration_directory: Path
    driver: str
    driver_adapter: Union[str, Any]
    dsn: str
    logger: logging.Logger
    aiosqlembic_queries: Queries
    current_version: Version


pass_settings = click.make_pass_decorator(AiosqlembicContext)


@click.group()
@click.option(
    "--version", is_flag=True, callback=print_version, expose_value=False, is_eager=True
)
@click.option("--debug/--no-debug", default=False)
@click.option(
    "-m",
    "--migration-directory",
    default=Path("."),
    type=click.Path(
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
    ),
)
@click.argument("driver", required=True, type=click.Choice(["asyncpg", "aiosqlite"]))
@click.argument("dsn", required=True)
@click.pass_context
async def cli(
    ctx: Context, debug: bool, migration_directory: Path, driver: str, dsn: str
) -> None:
    """
    Main entry point for the application
    :param ctx: the application context, no need to worry about it
    :param debug: a flag to log debug statuements, if set logging is set to DEBUG, if not it is set to INFO
    :param migration_directory: the directory where you want to store your migrations
    :param driver: can be asyncpg or aiosqlite
    :param dsn: the dsn of your database, a filepath in case of sqlite
    :return:
    """
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    try:
        driver_adapter = _ADAPTERS[driver]
    except KeyError:
        raise ValueError(f"Encountered unregistered driver_adapter: {driver_adapter}")

    aiosqlembic_queries = aiosql.from_path(
        Path(__file__).parent / "sql/aiosqlembic" / driver,
        driver,
        record_classes={
            "ConDB": ConDB,
            "ConDBTable": ConDBTable,
            "Version": Version,
            "MigrateRevision": MigrateRevision,
        },
    )

    try:
        async with get_connection(driver, dsn) as connection:
            conn_db: ConDB = await aiosqlembic_queries.check_connection(connection)
            if not conn_db.exists:
                logger.debug(f"Cant connect to dsn: {dsn}")
                click.Abort("Check dsn")

            conn_db_table = await aiosqlembic_queries.check_aiosqlembic(connection)
            if not conn_db_table.exists:
                logger.debug("aiosqlembic table doesnt exist")
                await aiosqlembic_queries.create_schema(connection)

                applied = True if driver == "asyncpg" else 1
                inserted = await aiosqlembic_queries.insert_revision(
                    connection, version_id=0, is_applied=applied
                )
                logger.debug(f"inserted: {inserted}")
            else:
                click.echo(f"Connected")

            current_version = await aiosqlembic_queries.get_version(connection)
            ctx.obj = AiosqlembicContext(
                debug=debug,
                migration_directory=Path(migration_directory),
                driver=driver,
                dsn=dsn,
                logger=logger,
                driver_adapter=driver_adapter,
                aiosqlembic_queries=aiosqlembic_queries,
                current_version=current_version,
            )
    except Exception as e:
        logger.error(e)
        raise e
    finally:
        pass


@cli.command()
@pass_settings
async def status(settings: AiosqlembicContext) -> None:
    logger.debug(
        f"STATUS with debug: {settings.debug} - migrations: {settings.migration_directory} - driver: {settings.driver} - dsn: {settings.dsn}"
    )
    revision_tree = get_revision_tree(
        settings.migration_directory, MIN_VERSION, MAX_VERSION
    )
    statusline = (
        "+" * 10
        + " time "
        + "+" * 10
        + " " * 4
        + "+" * 7
        + " applied at "
        + "+" * 7
        + " " * 4
        + "+" * 10
        + " revision file "
        + "+" * 10
    )
    click.echo(statusline)
    async with get_connection(settings.driver, settings.dsn) as connection:
        for revision in revision_tree:
            migrate_revision: MigrateRevision = await settings.aiosqlembic_queries.migrate_revision(
                connection, version_id=revision.version_id
            )

            if migrate_revision is not None and migrate_revision.is_applied:
                applied_ts = migrate_revision.tstamp.isoformat()
            else:
                applied_ts = "pending"
            click.echo(
                f"{datetime.utcnow().isoformat():30s}{applied_ts:15s}    {revision.file}"
            )


@cli.command()
@click.option("--name", "-n", required=True)
@click.option("--auto", "-a", required=False, default=False, type=click.BOOL)
@pass_settings
async def create(settings: AiosqlembicContext, name: str, auto: bool) -> None:
    logger.debug(
        f"Create with debug: {settings.debug} - migrations: {settings.migration_directory} - driver: {settings.driver} - dsn: {settings.dsn}"
    )
    version_id = datetime.utcnow().strftime("%Y%m%d%H%M%S")
    logger.debug(version_id)
    nospacename = name.replace(" ", "_")
    expected_revision_file = (
        settings.migration_directory / f"{version_id}_{camel2snake(nospacename)}.sql"
    )
    logger.debug(expected_revision_file)
    expected_revision_file.touch()
    revision_template = "revision.sql"
    template = templateEnv.get_template(revision_template)

    if auto:
        click.echo(
            "Attempt at generating sql statements! Please review them once written!"
        )
        # we load in theorical db the current list of revisions up the latest one
        async with create_test_database(
            settings.dsn, "theorical_aiosqlembic"
        ) as theorical:
            try:
                # putting theorical up to date with revision files
                revision_queries: Queries = aiosql.from_path(
                    settings.migration_directory, settings.driver
                )
                up_queries = [
                    x for x in revision_queries.available_queries if "up" in x
                ]
                pool_theorical = await asyncpg.create_pool(theorical)
                if up_queries:
                    for up in up_queries:
                        await getattr(revision_queries, up)(pool_theorical)

                pool_current = await asyncpg.create_pool(settings.dsn)
                # up
                up_mig = await Amigration().create(pool_theorical, pool_current)
                up_mig.set_safety(False)
                up_mig.add_all_changes()
                logger.debug(up_mig.statements)
                up_ss = up_mig.sql
                upgrade_statements = up_ss
                # down
                down_mig = await Amigration().create(pool_current, pool_theorical)
                down_mig.set_safety(False)
                down_mig.add_all_changes()
                logger.debug(down_mig.statements)
                down_ss = down_mig.sql
                downgrade_statements = down_ss
            except Exception as e:
                logger.error(e)
                raise e
            finally:
                await pool_current.close()
                await pool_theorical.close()

    else:
        upgrade_statements = "SELECT 'upgrade sql query here';"
        downgrade_statements = "SELECT 'downgrade sql query here';"

    output = template.render(
        upgrade_statements=upgrade_statements,
        downgrade_statements=downgrade_statements,
    )
    expected_revision_file.write_text(output)
    click.echo(f"Created: {expected_revision_file}")


async def get_current(settings: AiosqlembicContext) -> Version:
    async with get_connection(settings.driver, settings.dsn) as connection:
        versions: List[Version] = await settings.aiosqlembic_queries.get_version(
            connection
        )
    for version in versions:
        if version.is_applied:
            return version
    return Version(version_id=-1, is_applied=False)


async def upTo(settings: AiosqlembicContext, version_id: int) -> None:
    revision_tree = get_revision_tree(
        settings.migration_directory, MIN_VERSION, version_id
    )
    for _ in revision_tree:
        current: Version = await get_current(settings)
        next_revision = get_next_revision(revision_tree, current.version_id)
        if next_revision is not None:
            revision_query = aiosql.from_path(next_revision.file, settings.driver)
            try:
                async with get_connection(settings.driver, settings.dsn) as connection:
                    if settings.driver == "asyncpg":
                        async with connection.transaction():
                            result = await revision_query.up(connection)
                            logger.debug(f"up revision result: {result}")
                            await settings.aiosqlembic_queries.insert_revision(
                                connection,
                                version_id=next_revision.version_id,
                                is_applied=True,
                            )
                    elif settings.driver == "aiosqlite":
                        result = await revision_query.up(connection)
                        logger.debug(f"up revision result: {result}")
                        await settings.aiosqlembic_queries.insert_revision(
                            connection,
                            version_id=next_revision.version_id,
                            is_applied=1,
                        )
            except Exception as e:
                raise e
            click.echo(f"Applied revision: {next_revision.version_id}")


async def downTo(settings: AiosqlembicContext, version_id: int) -> None:
    revision_tree = get_revision_tree(
        settings.migration_directory, MIN_VERSION, MAX_VERSION
    )
    current: Version = await get_current(settings)
    current_revision = get_current_revision(revision_tree, current.version_id)
    if current_revision is not None:
        try:
            revision_query = aiosql.from_path(current_revision.file, settings.driver)
            async with get_connection(settings.driver, settings.dsn) as connection:
                if settings.driver == "asyncpg":
                    async with connection.transaction():
                        result = await revision_query.down(connection)
                        logger.debug(f"up revision result: {result}")
                        await settings.aiosqlembic_queries.delete_revision(
                            version_id=current_revision.version_id,
                        )
                elif settings.driver == "aiosqlite":
                    result = await revision_query.down(connection)
                    logger.debug(f"up revision result: {result}")
                    await settings.aiosqlembic_queries.delete_revision(
                        version_id=current_revision.version_id,
                    )
        except Exception as e:
            raise e


@cli.command()
@pass_settings
async def up(settings: AiosqlembicContext) -> None:
    logger.debug(
        f"UP with debug: {settings.debug} - migrations: {settings.migration_directory} - driver: {settings.driver} - dsn: {settings.dsn}"
    )
    await upTo(settings, MAX_VERSION)


@cli.command()
@pass_settings
async def down(settings: AiosqlembicContext) -> None:
    logger.debug(
        f"DOWN with debug: {settings.debug} - migrations: {settings.migration_directory} - driver: {settings.driver} - dsn: {settings.dsn}"
    )
    await downTo(settings, MAX_VERSION)


def main() -> None:
    cli(_anyio_backend="asyncio")


if __name__ == "__main__":
    main()



from typing import List, Optional

from pydantic.main import BaseModel


class Schema(BaseModel):
    schemaname: str


class Enum(BaseModel):
    schemaname: str
    name: str
    elements: List[str]


class Relation(BaseModel):
    attname: str
    collation: Optional[str]
    comment: Optional[str]
    datatype: str
    datatypestring: str
    defaultdef: Optional[str]
    definition: Optional[str]
    enum_name: Optional[str]
    enum_schema: Optional[str]
    forcerowsecurity: bool
    is_enum: bool
    name: str
    not_null: bool
    oid: int
    parent_table: Optional[str]
    partition_def: Optional[str]
    position_number: int
    relationtype: str
    rowsecurity: str
    schemaname: str


class Index(BaseModel):
    definition: str
    extension_oid: Optional[int]
    is_clustered: bool
    is_exclusion: bool
    is_immediate: bool
    is_pk: bool
    is_unique: bool
    key_collations: Optional[List[int]]  # oidvector
    key_columns: str
    key_expressions: Optional[str]
    key_options: Optional[List[int]]  # int2vector
    name: str
    num_att: int
    oid: int
    partial_predicate: Optional[str]
    schemaname: str
    table_name: str


class Sequence(BaseModel):
    schemaname: str
    name: str


class Constraint(BaseModel):
    constraint_type: str
    definition: str
    extension_oid: Optional[int]
    index: Optional[str]
    name: str
    schemaname: str
    table_name: str


class Extension(BaseModel):
    name: str
    oid: int
    schemaname: str
    version: str


class Function(BaseModel):
    comment: str
    data_type: str
    definition: str
    extension_oid: int
    full_definition: str
    has_user_defined_returntype: bool
    identity_arguments: str
    language: str
    name: str
    oid: int
    parameter_default: str
    parameter_mode: str
    parameter_name: str
    position_number: int
    result_string: str
    returntype: str
    schemaname: str
    security_type: str
    strictness: str
    volatility: str


class Dep(BaseModel):
    identity_arguments: str
    identity_arguments_dependent_on: str
    name: str
    name_dependent_on: str
    objid: int
    objid_dependent_on: int
    schema_dependent_on: str
    schemaname: str


class Privilege(BaseModel):
    name: str
    object_type: str
    privilege: str
    schemaname: str
    user: str


class Trigger(BaseModel):
    enabled: str
    full_definition: str
    name: str
    proc_name: str
    proc_schema: str
    schemaname: str
    table_name: str


class Collation(BaseModel):
    encoding: int
    lc_collate: str
    lc_ctype: str
    name: str
    provider: str
    schemaname: str
    version: str


class RlsPolicy(BaseModel):
    commandtype: str
    name: str
    permissive: bool
    qual: Optional[str]
    roles: Optional[List[str]]
    schemaname: str
    table_name: str
    withcheck: Optional[str]


class Type(BaseModel):
    columns: str
    columnsjson: str  # json
    description: str
    internal_name: str
    name: str
    schemaname: str
    size: str


class Domain(BaseModel):
    check: str
    collation: str
    constraint_name: str
    data_type: str
    default: str
    name: str
    not_null: bool
    schemaname: str



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiosqlembic', 'aiosqlembic.models']

package_data = \
{'': ['*'],
 'aiosqlembic': ['sql/aiosqlembic/aiosqlite/*',
                 'sql/aiosqlembic/asyncpg/*',
                 'sql/metadata/*',
                 'templates/*']}

install_requires = \
['aiosql>=3.0.0,<4.0.0',
 'aiosqlite>=0.11.0,<0.12.0',
 'anyio>=1.2.3,<2.0.0',
 'asyncclick>=7.0.9,<8.0.0',
 'asyncpg>=0.20.1,<0.21.0',
 'jinja2>=2.11.0,<3.0.0',
 'migra>=1.0.1575954682,<2.0.0',
 'pydantic>=1.3,<2.0',
 'trio>=0.13.0,<0.14.0']

entry_points = \
{'console_scripts': ['aiosqlembic = aiosqlembic.main:main']}

setup_kwargs = {
    'name': 'aiosqlembic',
    'version': '0.2.1',
    'description': 'Migrations powered by aiosql',
    'long_description': "Readme\n======\n\n.. image:: https://gitlab.com/euri10/aiosqlembic/badges/master/pipeline.svg\n    :target: https://gitlab.com/euri10/aiosqlembic/-/commits/master\n    :alt: pipeline status\n.. image:: https://gitlab.com/euri10/aiosqlembic/badges/master/coverage.svg\n    :target: https://gitlab.com/euri10/aiosqlembic/-/commits/master\n    :alt: coverage report\n.. image:: https://readthedocs.org/projects/aiosqlembic/badge/?version=latest\n    :target: https://aiosqlembic.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\nAiosqlembic aims at running database migrations powered by the awesome `aiosql <https://github.com/nackjicholson/aiosql>`_\n\nIt currently uses the also excellent `migra <https://github.com/djrobstep/migral>`_\n\nIt is in development and likely to break\n\nInstallation\n------------\n\nRun :code:`pip install aiosqlembic`\n\nDocumentation\n-------------\n\nIt's here `here <https://aiosqlembic.readthedocs.io/en/latest/index.html>`_\n\nUsage\n-----\n\nRun :code:`aiosqlembic --help`\n",
    'author': 'euri10',
    'author_email': 'benoit.barthelet@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/euri10/aiosqlembic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
